from __future__ import annotations

import inspect
from typing import Any, Optional

import numpy as np

from .base import TessellationBase as _TessType
from .dim2 import Tessellation2D as _Tess2D
from .dim3 import Tessellation3D as _Tess3D
from .exceptions import LowDimensionalityException
from .generic import TessellationGeneric as _TessND


def Tessellation(
    orbit_or_point_array: Any,
    dims_for_orbit: Optional[tuple[str]] = None,
    *,
    incremental: bool = True,
    qhull_options: Optional[str] = None,
    axis_ratio: float = 10,
    normalization_routine: str = "default",
    verbosity: int = 0,
) -> _TessType:
    points = orbit = orbit_or_point_array

    pkg = None
    module = inspect.getmodule(orbit)
    if module:
        pkg, *_ = module.__name__.partition(".")

    if pkg in ("galpy", "gala", "agama"):
        orbit = orbit_or_point_array
        if dims_for_orbit is None:
            raise TypeError("Orbit dimensions must be supplied if tessellating an orbit object")
        if pkg == "galpy":
            axes = [getattr(orbit, dim) for dim in dims_for_orbit]
            points = np.array([ax(orbit.t) for ax in axes]).T
        elif pkg == "gala":
            points = np.array([getattr(orbit, dim).value for dim in dims_for_orbit]).T
        elif pkg == "agama":
            raise NotImplementedError()
    else:
        if dims_for_orbit is not None:
            raise ValueError("Orbit dimensions must only be passed with an orbit object")
        points = np.array(points, dtype=float)

    if points.ndim != 2:
        raise ValueError(
            "Tessellation can only be performed on a 2D array of shape (npoints, ndim)"
        )

    init_args = dict(
        points=points,
        incremental=incremental,
        qhull_options=qhull_options,
        axis_ratio=axis_ratio,
        normalization_routine=normalization_routine,
        verbosity=verbosity,
    )

    # NOTE: ndim is not the same as points.ndim
    *_, ndim = points.shape
    if ndim == 2:
        return _Tess2D(**init_args)
    elif ndim == 3:
        return _Tess3D(**init_args)
    elif ndim > 3:
        return _TessND(**init_args)
    else:
        raise LowDimensionalityException(
            "Unsupported dimensionality, requires dimension 2 or greater"
        )



import inspect
import math
from importlib.metadata import version

import numpy as np
import pytest

import tessellation
from tessellation import Tessellation
from tessellation.exceptions import LowDimensionalityException

rng = np.random.default_rng(0)


class TestBasics:
    def test_version(self):
        assert version("orbit-tessellation") == tessellation.__version__


class TestInsufficientDimensionality:
    @pytest.mark.parametrize(
        "points",
        [
            True,
            False,
            0,
            1,
            -1,
            0.0,
            1.5,
            1.482e-100,
        ],
    )
    def test_zero_dimensional_numeric_builtins(self, points):
        arr = np.array(points)
        assert arr.size == 1
        assert arr.ndim == 0
        assert arr.shape == ()
        with pytest.raises(ValueError):
            Tessellation(points)
        with pytest.raises(ValueError):
            Tessellation(arr)

    @pytest.mark.parametrize(
        "points",
        [
            {},
            {"a": 1},
            {"a": 1, "b": 2},
            set(),
            {0},
            {0, 1},
            {0, 1, 2, 3},
        ],
    )
    def test_zero_dimensional_container_builtins(self, points):
        arr = np.array(points)
        assert arr.size == 1
        assert arr.ndim == 0
        assert arr.shape == ()
        with pytest.raises(TypeError):
            Tessellation(points)
        with pytest.raises(TypeError):
            Tessellation(arr)

    @pytest.mark.filterwarnings("ignore:Casting complex values")
    @pytest.mark.parametrize(
        "points,exc1,exc2",
        [
            (1j, TypeError, ValueError),
            ("", ValueError, ValueError),
            ("a", ValueError, ValueError),
            ("asdf", ValueError, ValueError),
        ],
    )
    def test_zero_dimensional_noncasting_builtins(self, points, exc1, exc2):
        arr = np.array(points)
        assert arr.size == 1
        assert arr.ndim == 0
        assert arr.shape == ()
        with pytest.raises(exc1):
            Tessellation(points)
        with pytest.raises(exc2):
            Tessellation(arr)

    @pytest.mark.parametrize(
        "points",
        [
            [],
            [0],
            [0, 1],
            [0, 1, 2],
            [0, 1, 2, 3],
            (),
            (0,),
            (0, 1),
            (0, 1, 2),
            (0, 1, 2, 3),
        ],
    )
    def test_one_dimensional_container_builtins(self, points):
        arr = np.array(points)
        assert arr.ndim == 1
        with pytest.raises(ValueError):
            Tessellation(points)
        with pytest.raises(ValueError):
            Tessellation(arr)

    @pytest.mark.parametrize(
        "points",
        [
            [[], []],
            [[0], [1]],
            ([], []),
            ([0], [1]),
            [(), ()],
            [(0,), (1,)],
            ((), ()),
            ((0,), (1,)),
        ],
    )
    def test_two_dimensional_container_builtins(self, points):
        arr = np.array(points)
        print(points, arr)
        assert arr.ndim == 2
        with pytest.raises(LowDimensionalityException):
            Tessellation(points)
        with pytest.raises(LowDimensionalityException):
            Tessellation(arr)

    @pytest.mark.parametrize(
        "points",
        [
            [[], [0]],
            [[], [], [1], []],
            [[0], [1, 2]],
            [[0, 1], [2]],
        ],
    )
    def test_two_dimensional_container_builtins_inhomogeneity(self, points):
        with pytest.raises(ValueError, match="inhomogeneous"):
            Tessellation(points)

    @pytest.mark.parametrize(
        "points",
        [
            rng.random((0, 0)),
            rng.random((1, 0)),
            rng.random((2, 0)),
            rng.random((0, 1)),
            rng.random((1, 1)),
            rng.random((2, 1)),
        ],
    )
    def test_ndarray(self, points):
        with pytest.raises(LowDimensionalityException):
            Tessellation(points)


class TestDegeneracies:
    @pytest.mark.parametrize(
        "points",
        [
            rng.random((0, 2)),
            rng.random((0, 3)),
            rng.random((0, 4)),
        ],
    )
    def test_empty_point_array(self, points):
        with pytest.raises(ValueError, match="No points given"):
            Tessellation(points)

    @pytest.mark.parametrize(
        "points",
        [
            rng.random((1, 2)),
            rng.random((2, 2)),
            rng.random((3, 2)),
        ],
    )
    def test_insufficient_points_in_two_dimensions(self, points):
        with pytest.warns(UserWarning, match="degenerate"):
            tess = Tessellation(points)
        assert tess.measure == 0.0

    @pytest.mark.parametrize(
        "points",
        [
            rng.random((1, 3)),
            rng.random((2, 3)),
            rng.random((3, 3)),
            rng.random((4, 3)),
        ],
    )
    def test_insufficient_points_in_three_dimensions(self, points):
        with pytest.warns(UserWarning, match="degenerate"):
            tess = Tessellation(points)
        assert tess.measure == 0.0

    @pytest.mark.parametrize(
        "points",
        [
            rng.random((1, 4)),
            rng.random((2, 4)),
            rng.random((3, 4)),
            rng.random((4, 4)),
            rng.random((5, 4)),
        ],
    )
    def test_insufficient_points_in_four_dimensions(self, points):
        with pytest.warns(UserWarning, match="degenerate"):
            tess = Tessellation(points)
        assert tess.measure == 0.0

    @pytest.mark.parametrize(
        "points",
        [
            np.array(
                [
                    np.zeros(100),
                    np.zeros(100),
                ]
            ).T,
            np.array(
                [
                    np.zeros(100),
                    rng.random(100),
                ]
            ).T,
            np.array(
                [
                    rng.random(100),
                    np.zeros(100),
                ]
            ).T,
            np.array(
                [
                    np.arange(100),
                    np.arange(100),
                ]
            ).T,
            np.array(
                [
                    np.arange(100),
                    np.arange(100) * 2,
                ]
            ).T,
            np.array(
                [
                    (x := rng.random(100)),
                    3 * x - 4,
                ]
            ).T,
        ],
    )
    def test_degeneracy_in_two_dimensions(self, points):
        with pytest.warns(UserWarning, match="degenerate"):
            tess = Tessellation(points)
        assert tess.measure == 0.0

    @pytest.mark.parametrize(
        "points",
        [
            np.array(
                [
                    np.zeros(100),
                    np.zeros(100),
                    np.zeros(100),
                ]
            ).T,
            np.array(
                [
                    rng.random(100),
                    np.zeros(100),
                    np.zeros(100),
                ]
            ).T,
            np.array(
                [
                    np.zeros(100),
                    rng.random(100),
                    np.zeros(100),
                ]
            ).T,
            np.array(
                [
                    np.zeros(100),
                    np.zeros(100),
                    rng.random(100),
                ]
            ).T,
            np.array(
                [
                    rng.random(100),
                    rng.random(100),
                    np.zeros(100),
                ]
            ).T,
            np.array(
                [
                    rng.random(100),
                    np.zeros(100),
                    rng.random(100),
                ]
            ).T,
            np.array(
                [
                    np.zeros(100),
                    rng.random(100),
                    rng.random(100),
                ]
            ).T,
            np.array(
                [
                    np.arange(100),
                    np.arange(100),
                    np.arange(100),
                ]
            ).T,
            np.array(
                [
                    np.arange(100),
                    np.arange(100) * 2,
                    np.arange(100) * 3,
                ]
            ).T,
            np.array(
                [
                    (x := rng.random(100)),
                    5 - 7 * x,
                    4 + 3 * x,
                ]
            ).T,
            np.array(
                [
                    (x := rng.random(100)),
                    (y := rng.random(100)),
                    9 + 4 * x - 3 * y,
                ]
            ).T,
        ],
    )
    def test_degeneracy_in_three_dimensions(self, points):
        with pytest.warns(UserWarning, match="degenerate"):
            tess = Tessellation(points)
        assert tess.measure == 0.0

    @pytest.mark.parametrize(
        "points",
        [
            np.array(
                [
                    np.zeros(100),
                    np.zeros(100),
                    np.zeros(100),
                    np.zeros(100),
                ]
            ).T,
            np.array(
                [
                    rng.random(100),
                    np.zeros(100),
                    np.zeros(100),
                    np.zeros(100),
                ]
            ).T,
            np.array(
                [
                    np.zeros(100),
                    rng.random(100),
                    np.zeros(100),
                    np.zeros(100),
                ]
            ).T,
            np.array(
                [
                    np.zeros(100),
                    np.zeros(100),
                    rng.random(100),
                    np.zeros(100),
                ]
            ).T,
            np.array(
                [
                    np.zeros(100),
                    np.zeros(100),
                    np.zeros(100),
                    rng.random(100),
                ]
            ).T,
            np.array(
                [
                    rng.random(100),
                    rng.random(100),
                    np.zeros(100),
                    np.zeros(100),
                ]
            ).T,
            np.array(
                [
                    rng.random(100),
                    np.zeros(100),
                    rng.random(100),
                    np.zeros(100),
                ]
            ).T,
            np.array(
                [
                    rng.random(100),
                    np.zeros(100),
                    np.zeros(100),
                    rng.random(100),
                ]
            ).T,
            np.array(
                [
                    np.zeros(100),
                    rng.random(100),
                    rng.random(100),
                    np.zeros(100),
                ]
            ).T,
            np.array(
                [
                    np.zeros(100),
                    rng.random(100),
                    np.zeros(100),
                    rng.random(100),
                ]
            ).T,
            np.array(
                [
                    np.zeros(100),
                    np.zeros(100),
                    rng.random(100),
                    rng.random(100),
                ]
            ).T,
            np.array(
                [
                    rng.random(100),
                    rng.random(100),
                    rng.random(100),
                    np.zeros(100),
                ]
            ).T,
            np.array(
                [
                    rng.random(100),
                    rng.random(100),
                    np.zeros(100),
                    rng.random(100),
                ]
            ).T,
            np.array(
                [
                    rng.random(100),
                    np.zeros(100),
                    rng.random(100),
                    rng.random(100),
                ]
            ).T,
            np.array(
                [
                    np.zeros(100),
                    rng.random(100),
                    rng.random(100),
                    rng.random(100),
                ]
            ).T,
            np.array(
                [
                    np.arange(100),
                    np.arange(100),
                    np.arange(100),
                    np.arange(100),
                ]
            ).T,
            np.array(
                [
                    np.arange(100),
                    np.arange(100) * 2,
                    np.arange(100) * 3,
                    np.arange(100) * 4,
                ]
            ).T,
            np.array(
                [
                    (x := rng.random(100)),
                    8 * x - 9,
                    7 * x + 5,
                    2 - 3 * x,
                ]
            ).T,
            np.array(
                [
                    (x := rng.random(100)),
                    (y := rng.random(100)),
                    3 + 4 * x - 6 * y,
                    2 - 9 * x - 8 * y,
                ]
            ).T,
            np.array(
                [
                    (x := rng.random(100)),
                    (y := rng.random(100)),
                    3 + 4 * x - 6 * y,
                    2 - 9 * x - 8 * y,
                ]
            ).T,
            np.array(
                [
                    (x := rng.random(100)),
                    (y := rng.random(100)),
                    (z := rng.random(100)),
                    4 * x - 4 * y + 3 * z - 18,
                ]
            ).T,
        ],
    )
    def test_degeneracy_in_four_dimensions(self, points):
        with pytest.warns(UserWarning, match="degenerate"):
            tess = Tessellation(points)
        assert tess.measure == 0.0


class TestMeasure:
    @pytest.mark.parametrize("points", rng.normal(-1, 1, (20, 100, 2)))
    def test_normalized_range(self, points):
        for axis_ratio in (0, 5, 10, 15, 20, 50, 100, np.inf):
            tess = Tessellation(points, axis_ratio=axis_ratio)
            assert 0 <= tess.measure <= 1

    @pytest.mark.parametrize(
        "points,expected_measure",
        [
            ([[1, 0], [0, 1], [-1, 0], [0, 0]], 1 / np.pi),
            ([[1, 0], [0, 1], [-1, 0], [0, -1]], 2 / np.pi),
            ([[0, 0], [0, 1], [1, 0], [1, 1]], 1 / (2 * np.pi)),
        ],
    )
    def test_manual_point_sets_in_2D(self, points, expected_measure):
        tess = Tessellation(points, axis_ratio=np.inf, qhull_options="Qz", incremental=False)
        assert math.isclose(tess.measure, expected_measure)

    @pytest.mark.parametrize(
        "points,expected_measure",
        [
            (
                [
                    [1, 0, 0],
                    [-1, 0, 0],
                    [0, 1, 0],
                    [0, -1, 0],
                    [0, 0, 1],
                    [0, 0, -1],
                ],
                1,
            )
        ],
    )
    def test_manual_point_sets_in_3D(self, points, expected_measure):
        tess = Tessellation(points, axis_ratio=np.inf, qhull_options="Qz", incremental=False)
        assert math.isclose(tess.measure, expected_measure)

    @pytest.mark.parametrize(
        "points,expected_measure",
        [
            (rng.normal(size=(100, 2)), 0.45906549045400247),
            (rng.normal(size=(100, 2)), 0.3943940252036827),
            (rng.normal(size=(100, 2)), 0.4336073562797213),
            (rng.normal(size=(100, 2)), 0.5617838554666587),
            (rng.normal(size=(100, 2)), 0.5985426693438319),
            (rng.normal(size=(100, 2)), 0.5123637204843497),
            (rng.normal(size=(100, 2)), 0.3225978521940026),
            (rng.normal(size=(100, 2)), 0.4143084419365352),
            (rng.normal(size=(100, 2)), 0.5987423967178952),
            (rng.normal(size=(100, 2)), 0.45548485279657114),
        ],
    )
    def test_pseudorandom_point_sets_in_2D(self, points, expected_measure):
        tess = Tessellation(points)
        assert tess.measure == expected_measure

    @pytest.mark.parametrize(
        "points,expected_measure",
        [
            (rng.normal(size=(100, 3)), 0.5919362630015677),
            (rng.normal(size=(100, 3)), 0.4964502307751198),
            (rng.normal(size=(100, 3)), 0.5315301767521167),
            (rng.normal(size=(100, 3)), 0.5627029309700805),
            (rng.normal(size=(100, 3)), 0.6854798548331412),
            (rng.normal(size=(100, 3)), 0.4686263346093926),
            (rng.normal(size=(100, 3)), 0.6370942709595918),
            (rng.normal(size=(100, 3)), 0.4930328558372991),
            (rng.normal(size=(100, 3)), 0.5546090138460925),
            (rng.normal(size=(100, 3)), 0.5457828455144428),
        ],
    )
    def test_pseudorandom_point_sets_in_3D(self, points, expected_measure):
        tess = Tessellation(points)
        assert tess.measure == expected_measure



from __future__ import annotations

from itertools import combinations

import numpy as np
from scipy import linalg, spatial

from .base import TessellationBase


class TessellationGeneric(TessellationBase):
    @staticmethod
    def simplex_sides(*vertices):
        # combinatoric approach - not gonna use this in practice, hopefully
        return [v2 - v1 for v1, v2 in combinations(vertices, 2)]

    @staticmethod
    def simplex_measure(*vertices):
        # general simplex volume - not gonna use this in practice, hopefully
        first, *rest = vertices
        dim = len(first)
        mat = [v - first for v in rest]
        return linalg.det(mat) / np.math.factorial(dim)

    class Normalization(TessellationBase):
        def nsphere_approx(self) -> float:
            # stirling's approximation - not gonna use this in practice, hopefully
            d = self.points.shape[1]
            r = linalg.norm(self.points, axis=1)
            return 1 / np.sqrt(d * np.pi) * (2 * np.pi * np.e / d) ** (d / 2) * np.max(r) ** d

        def convexhull(self) -> float:
            # breaks for co-rotation cases
            points = self.tri.convex_hull
            return spatial.ConvexHull(points).volume

        def convexhull_rot4(self):
            x, y, *rest = self.points.T
            r000 = np.array([+x, +y, *rest]).T
            r090 = np.array([-y, +x, *rest]).T
            r180 = np.array([-x, -y, *rest]).T
            r270 = np.array([+y, -x, *rest]).T

            points = np.array([*r000, *r090, *r180, *r270])
            hull = spatial.ConvexHull(points)
            return hull.volume

        default = convexhull_rot4



import numpy as np

try:
    import matplotlib.pyplot as plt
    import mpl_toolkits.mplot3d as a3

    PLOTTING = True
except ImportError:
    PLOTTING = False

from scipy import linalg, spatial

from .base import TessellationBase


class Tessellation3D(TessellationBase):
    @staticmethod
    def simplex_sides(*vertices):
        v1, v2, v3, v4 = vertices
        return [
            linalg.norm(v2 - v1),
            linalg.norm(v3 - v1),
            linalg.norm(v4 - v1),
            linalg.norm(v3 - v2),
            linalg.norm(v4 - v2),
            linalg.norm(v4 - v3),
        ]

    @staticmethod
    def simplex_measure(*vertices):
        (x1, y1, z1), (x2, y2, z2), (x3, y3, z3), (x4, y4, z4) = vertices
        a1 = (x2 - x1) * ((y3 - y1) * (z4 - z1) - (y4 - y1) * (z3 - z1))
        a2 = (x3 - x1) * ((y4 - y1) * (z2 - z1) - (y2 - y1) * (z4 - z1))
        a3 = (x4 - x1) * ((y2 - y1) * (z3 - z1) - (y3 - y1) * (z2 - z1))
        return abs(a1 + a2 + a3) / 6

    class Normalization:
        def sphere(self):
            r = linalg.norm(self.points, axis=1)
            return 4 / 3 * np.pi * np.max(r) ** 3

        def cylinder(self):
            x, y, z = self.points.T
            return np.pi * np.max(x**2 + y**2) * (np.max(z) - np.min(z))

        def Rz_convexhull(self):
            x, y, z = self.points.T
            R = np.sqrt(x**2 + y**2)
            points = np.array([R, z]).T
            points = np.array([*points, [0, max(z)], [0, min(z)]])
            hull = spatial.ConvexHull(points)

            poly = points[hull.vertices]
            start = poly[0]
            poly = poly[1:] - start

            # NOTE: requires "pairwise" from itertools (recent python version)
            from itertools import pairwise

            centroids = [start + (t1 + t2) / 3 for t1, t2 in pairwise(poly)]
            areas = [np.linalg.det([t1, t2]) for t1, t2 in pairwise(poly)]

            centroid = np.sum([a * c for c, a in zip(centroids, areas)], axis=0)
            return 2 * np.pi * np.linalg.norm(centroid[:2])

        def convexhull(self):
            hull = spatial.ConvexHull(self.points)
            return hull.volume

        def convexhull_rot4(self):
            x, y, z = self.points.T
            r000 = np.array([+x, +y, z]).T
            r090 = np.array([-y, +x, z]).T
            r180 = np.array([-x, -y, z]).T
            r270 = np.array([+y, -x, z]).T

            points = np.array([*r000, *r090, *r180, *r270])
            hull = spatial.ConvexHull(points)
            return hull.volume

        default = convexhull_rot4

    @property
    def volume(self):
        return self.measure

    def plot(
        self,
        plot_included=True,
        plot_removed=False,
        plot_points=True,
        verbosity=1,
        ax=None,
        show=True,
    ):
        """
        Plot the triangulation - trimmed triangles are drawn in red.
        """
        if not PLOTTING:
            raise ImportError("This method requires matplotlib")
        if self.tri is None:
            raise RuntimeError("Tessellation failed; cannot produce tessellation plot")

        if not ax:
            fig = plt.figure()
            ax = fig.add_subplot(projection="3d")
        X, Y, Z = self.points.T

        included_edges = set()
        excluded_edges = set()
        for simplex, included in zip(self.tri.simplices, self.mask):
            i1, i2, i3, i4 = simplex
            edges = included_edges if included else excluded_edges
            for tuplet in (
                (i1, i2),
                (i1, i3),
                (i1, i4),
                (i2, i3),
                (i2, i4),
                (i3, i4),
            ):
                edges.add(tuplet)

        if plot_removed and plot_included:
            inex_edges = included_edges & excluded_edges
            included_edges -= inex_edges
            excluded_edges -= inex_edges

            inex_lines = [(self.points[i1], self.points[i2]) for i1, i2 in inex_edges]
            if verbosity:
                print(self.__class__.__name__, "plotting in/ex edges (blue):", len(inex_lines))

            line_collection = a3.art3d.Poly3DCollection(inex_lines)
            line_collection.set_edgecolor("blue")
            line_collection.set_linewidths(0.2)
            ax.add_collection3d(line_collection)

        if plot_removed:
            excluded_lines = [(self.points[i1], self.points[i2]) for i1, i2 in excluded_edges]
            if verbosity:
                print(
                    self.__class__.__name__, "plotting excluded edges (red):", len(excluded_lines)
                )

            line_collection = a3.art3d.Poly3DCollection(excluded_lines)
            line_collection.set_edgecolor("red")
            line_collection.set_linewidths(0.2)
            ax.add_collection3d(line_collection)

        if plot_included:
            included_lines = [(self.points[i1], self.points[i2]) for i1, i2 in included_edges]
            if verbosity:
                print(
                    self.__class__.__name__, "plotting included edges (green):", len(included_lines)
                )

            line_collection = a3.art3d.Poly3DCollection(included_lines)
            line_collection.set_edgecolor("green")
            line_collection.set_linewidths(0.2)
            ax.add_collection3d(line_collection)

        if plot_points:
            ax.scatter(X, Y, Z, marker=".", color="black")
            if verbosity:
                print(self.__class__.__name__, "plotting points:", len(X))

        ax_lim = 1.1 * max(max(X), max(Y), max(Z))
        ax_lim = (-ax_lim, ax_lim)
        ax.set(xlim=ax_lim, ylim=ax_lim, zlim=ax_lim)

        if show:
            plt.show()
        return ax



from __future__ import annotations

import warnings
from abc import abstractstaticmethod
from collections import defaultdict
from typing import Any, Callable, Mapping, Optional, Union

import numpy as np
from scipy import spatial

MISSING = object()


def MISSING_METHOD() -> None:
    pass


class FailedDelaunay:
    def __init__(self):
        self.nsimplex = 0
        self.simplices = [[0]]
        self.convex_hull = [[0]]


class TessellationBase:
    __slots__ = (
        "points",
        "normalizations",
        "tri",
        "tracers",
        "mask",
        "normalization_const",
        "measure",
    )

    def __init__(
        self,
        points: np.ndarray,
        *,
        incremental: bool = True,
        qhull_options: Optional[str] = None,
        axis_ratio: float = 10,
        normalization_routine: str = "default",
        verbosity: int = 0,
    ) -> None:
        self.points: np.ndarray = points
        self._normalizations: Mapping[str, Callable]
        self.tri: Union[spatial.Delaunay, FailedDelaunay]
        self.tracers: Mapping[str, np.ndarray]
        self.mask: Optional[np.ndarray]
        self.normalization_const: float
        self.measure: Optional[float]
        verbosity = verbosity

        normalization_class = getattr(self, "Normalization", type("Normalization", (), {}))
        self._normalizations = dict(vars(normalization_class))
        del self._normalizations["__module__"]
        del self._normalizations["__dict__"]
        del self._normalizations["__weakref__"]
        del self._normalizations["__doc__"]

        try:
            r = self._compute_delaunay(incremental, qhull_options)
            if verbosity:
                if self.tri:
                    print(f"{self.__class__.__name__}: Delaunay tessellation computed.")
                else:
                    print(f"{self.__class__.__name__}: Delaunay tessellation failed. {r}")

            self._compute_tracers()
            if verbosity:
                print(f"{self.__class__.__name__}: Simplex sides and measures computed.")

            self._compute_trimming(axis_ratio)
            if verbosity:
                print(f"{self.__class__.__name__}: Tessellation trimming computed.")

            self._compute_normalization(normalization_routine)
            if verbosity:
                print(f"{self.__class__.__name__}: Normalization computed.")

            self.measure = np.sum(self.tracers["measure"][self.mask]) / self.normalization_const
        except AttributeError:
            warnings.warn(
                "Point set appears to be degenerate or co-spherical. "
                "Consider using a different dimensionality for the degenerate case."
            )

    def __repr__(self):
        status = "bare"
        if self.tri is not None:
            status = "triangulated"
        if self.mask is not None:
            status = "trimmed"
        if self.measure is not None:
            status = f"measure={self.measure}"
        return f"{self.__class__.__name__}[{status}]"

    def __setattr__(self, name: str, value: Any) -> None:
        existing_value = getattr(self, name, MISSING)
        if existing_value is not MISSING:
            raise AttributeError(
                f"Attribute {name} is already set to {existing_value}. "
                f"Please create a new Tessellation object for this operation. "
            )
        return super().__setattr__(name, value)

    def _compute_delaunay(
        self, incremental: bool, qhull_options: Optional[str]
    ) -> Optional[spatial.QhullError]:
        try:
            # Delaunay triangulation routine from SciPy:
            # https://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.Delaunay.html
            self.tri = spatial.Delaunay(
                self.points, incremental=incremental, qhull_options=qhull_options
            )
            return None
        except spatial.QhullError as e:
            # For degenerate cases, proceed with measure of 0
            self.tri = FailedDelaunay()
            # self.tri = None  # NOTE: might break when incremental=True
            self.tracers = defaultdict(lambda: np.zeros(self.tri.nsimplex))
            self.mask = None
            self.normalization_const = 1.0
            self.measure = 0.0
            return e

    def _compute_tracers(self) -> None:
        self.tracers = defaultdict(lambda: np.zeros(self.tri.nsimplex))
        for i, simplex in enumerate(self.tri.simplices):
            vertices = self.points[simplex]

            # store extreme side lengths (for trimming)
            side_lens = self.simplex_sides(*vertices)
            self.tracers["largest side"][i] = max(side_lens)
            self.tracers["smallest side"][i] = min(side_lens)

            # store simplex measures
            self.tracers["measure"][i] = self.simplex_measure(*vertices)

    def _compute_trimming(self, axis_ratio: float = 10) -> None:
        # self.mask = np.ones(self.tri.nsimplex, dtype=bool)

        # trim simplices with large sides
        threshold = axis_ratio * np.median(self.tracers["smallest side"])
        self.mask = self.tracers["largest side"] < threshold

    def _compute_normalization(self, routine: str) -> None:
        normalization_method: Callable = self._normalizations.get(routine, MISSING_METHOD)
        if normalization_method is MISSING_METHOD:
            message = f"Unrecognized normalization routine {routine}. "
            available = list(self._normalizations.keys())
            if available:
                message += f"Available normalizations are {available}"
            else:
                message += "No available normalizations for this class"
            raise ValueError(message)
        self.normalization_const = normalization_method(self)

    @abstractstaticmethod
    def simplex_sides(*vertices):
        pass

    @abstractstaticmethod
    def simplex_measure(*vertices):
        pass

    def plot(self, fig, plot_included=True, plot_removed=False, plot_points=True, verbosity=1):
        pass



import numpy as np
from scipy import linalg

try:
    import matplotlib.pyplot as plt

    PLOTTING = True
except ImportError:
    PLOTTING = False

from .base import TessellationBase


class Tessellation2D(TessellationBase):
    @staticmethod
    def simplex_sides(*vertices):
        v1, v2, v3 = vertices
        return [
            linalg.norm(v2 - v1),
            linalg.norm(v3 - v1),
            linalg.norm(v3 - v2),
        ]

    @staticmethod
    def simplex_measure(*vertices):
        (x1, y1), (x2, y2), (x3, y3) = vertices
        return abs((x2 - x1) * (y3 - y1) - (x3 - x1) * (y2 - y1)) / 2

    class Normalization:
        def circle(self):
            r = linalg.norm(self.points, axis=1)
            return np.pi * (max(r) ** 2)

        default = circle

    @property
    def area(self):
        return self.measure

    def plot(
        self,
        plot_included=True,
        plot_removed=False,
        plot_points=True,
        verbosity=1,
        ax=None,
        show=True,
    ):
        """
        Plot the triangulation - trimmed triangles are drawn in red.
        """
        if not PLOTTING:
            raise ImportError("This method requires matplotlib")
        if self.tri is None:
            raise RuntimeError("Tessellation failed; cannot produce tessellation plot")

        if not ax:
            fig = plt.figure()
            ax = fig.add_subplot()
        X, Y = self.points.T

        if plot_removed:
            plt.triplot(X, Y, self.tri.simplices, mask=self.mask, color="red")
            if verbosity:
                print(self.__class__.__name__, "plotting excluded edges (red):", len(self.mask))

        if plot_included:
            plt.triplot(X, Y, self.tri.simplices, mask=~self.mask, color="green")
            if verbosity:
                print(self.__class__.__name__, "plotting included edges (green):", len(~self.mask))

        if plot_points:
            plt.plot(X, Y, "k.", markersize=0.5)
            if verbosity:
                print(self.__class__.__name__, "plotting points:", len(X))

        if show:
            plt.show()
        return ax
